/* $Id: cpl_imagelist_io.c,v 1.63 2009/03/04 13:37:01 lbilbao Exp $
 *
 * This file is part of the ESO Common Pipeline Library
 * Copyright (C) 2001-2008 European Southern Observatory
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

/*
 * $Author: lbilbao $
 * $Date: 2009/03/04 13:37:01 $
 * $Revision: 1.63 $
 * $Name: cpl-5_0_0 $
 */

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

/*-----------------------------------------------------------------------------
                                Includes
 -----------------------------------------------------------------------------*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <math.h>
#include <assert.h>

#include <fitsio.h>

#include "cpl_tools.h"
#include "cpl_propertylist_impl.h"
#include "cpl_memory.h"
#include "cpl_imagelist_io.h"

#include "cpl_imagelist_defs.h"

/*-----------------------------------------------------------------------------
                                Defines
 -----------------------------------------------------------------------------*/

#define CPL_IMLIST_IO_LOAD_WIN  1

/*-----------------------------------------------------------------------------
                                Defines
 -----------------------------------------------------------------------------*/

static cpl_imagelist * cpl_imagelist_load_internal(const char  *, cpl_type,
						   int, int, int, int, int);

/*----------------------------------------------------------------------------*/
/**
 * @defgroup cpl_imagelist_io Image list IO routines
 *
 * This module provides IO functions to handle @em cpl_imagelist.
 *
 * @par Synopsis:
 * @code
 *   #include "cpl_imagelist_io.h"
 * @endcode
 */
/*----------------------------------------------------------------------------*/
/**@{*/

/*-----------------------------------------------------------------------------
                            Function codes
 -----------------------------------------------------------------------------*/

/*----------------------------------------------------------------------------*/
/**
  @brief    Create an empty imagelist
  @return   1 newly allocated cpl_imagelist
  @see      cpl_imagelist_set()

  The returned cpl_imagelist must be deallocated using cpl_imagelist_delete()

 */
/*----------------------------------------------------------------------------*/
cpl_imagelist * cpl_imagelist_new(void)
{
    return (cpl_imagelist *) cpl_calloc(1, sizeof(cpl_imagelist));
}

/*----------------------------------------------------------------------------*/
/**
  @brief    Load a FITS file extension into a list of images
  @param    filename    The FITS file name
  @param    type        Type of the images in the created image list
  @param    extnum      The extension number
  @return   The loaded list of images or NULL on error.
  @see      cpl_image_load()

  This function loads all the images of a specified extension (NAXIS=2
  or 3) in an image list.

  Type can be CPL_TYPE_DOUBLE, CPL_TYPE_FLOAT or CPL_TYPE_INT.
  The loaded images have an empty bad pixel map.

  The returned cpl_imagelist must be deallocated using cpl_imagelist_delete()

  Possible #_cpl_error_code_ set in this function:
  - CPL_ERROR_NULL_INPUT if (one of) the input pointer(s) is NULL
  - CPL_ERROR_ILLEGAL_INPUT if extnum is negative
  - CPL_ERROR_INVALID_TYPE if the passed type is not supported
  - CPL_ERROR_FILE_IO if extnum is bigger than the number of extensions in the
                    FITS file
  - CPL_ERROR_BAD_FILE_FORMAT if the file cannot be parsed
  - CPL_ERROR_DATA_NOT_FOUND if the datacannot be read from the file
 */
/*----------------------------------------------------------------------------*/
cpl_imagelist * cpl_imagelist_load(
        const char * filename,
        cpl_type     type,
        int          extnum)
{
    return cpl_imagelist_load_internal(filename, type, extnum, -1, -1, -1, -1);
}

/*----------------------------------------------------------------------------*/
/**
  @brief    Load images windows from a FITS file extension into an image list
  @param    filename    The FITS file name
  @param    type        Type of the images in the created image list
  @param    extnum      The extension number
  @param    llx         Specifies the window position
  @param    lly         Specifies the window position
  @param    urx         Specifies the window position
  @param    ury         Specifies the window position
  @return   The loaded list of image windows or NULL on error.
  @see      cpl_image_load_window()

  This function loads all the image windows of a specified extension in an
  image list.

  Type can be CPL_TYPE_DOUBLE, CPL_TYPE_FLOAT or CPL_TYPE_INT.

  The window position is specified starting with 1 to the number of
  pixel. Lower left pixel is (1, 1).

  The returned cpl_imagelist must be deallocated using cpl_imagelist_delete()

  Possible #_cpl_error_code_ set in this function:
  - CPL_ERROR_NULL_INPUT if (one of) the input pointer(s) is NULL
  - CPL_ERROR_ILLEGAL_INPUT if extnum is negative
  - CPL_ERROR_INVALID_TYPE if the passed type is not supported
  - CPL_ERROR_ACCESS_OUT_OF_RANGE if extnum is bigger than the number of
    extensions in the FITS file
  - CPL_ERROR_BAD_FILE_FORMAT if the file cannot be parsed
  - CPL_ERROR_DATA_NOT_FOUND if the datacannot be read from the file
 */
/*----------------------------------------------------------------------------*/
cpl_imagelist * cpl_imagelist_load_window(
        const char * filename,
        cpl_type     type,
        int          extnum,
        int          llx,
        int          lly,
        int          urx,
        int          ury)

{
    cpl_ensure(llx > 0 && lly > 0 && urx >= llx && ury >= lly,
	       CPL_ERROR_ILLEGAL_INPUT, NULL);

    return cpl_imagelist_load_internal(filename, type, extnum,
				       llx, lly, urx, ury);
}

/*----------------------------------------------------------------------------*/
/**
  @brief    Get the number of images in the imagelist
  @param    imlist    the list of image
  @return   The number of images or -1 on error

  Possible #_cpl_error_code_ set in this function:
  - CPL_ERROR_NULL_INPUT if (one of) the input pointer(s) is NULL
 */
/*----------------------------------------------------------------------------*/
int cpl_imagelist_get_size(const cpl_imagelist * imlist)
{
    cpl_ensure(imlist != NULL, CPL_ERROR_NULL_INPUT,
               -1);

    assert( imlist->ni >= 0 );

    return imlist->ni ;
}

/*----------------------------------------------------------------------------*/
/**
  @brief    Get an image from a list of images
  @param    imlist    the image list
  @param    inum    the image id (from 0 to number of images-1)
  @return   A pointer to the image or NULL in error case.

  The returned pointer refers to already allocated data.

  Possible #_cpl_error_code_ set in this function:
  - CPL_ERROR_NULL_INPUT if (one of) the input pointer(s) is NULL
  - CPL_ERROR_ACCESS_OUT_OF_RANGE if inum is bigger thant the list size
  - CPL_ERROR_ILLEGAL_INPUT if inum is negative
 */
/*----------------------------------------------------------------------------*/
cpl_image * cpl_imagelist_get(
        cpl_imagelist * imlist,
        int                   inum)
{

    cpl_ensure(imlist != NULL,    CPL_ERROR_NULL_INPUT, NULL);
    cpl_ensure(inum >= 0,         CPL_ERROR_ILLEGAL_INPUT, NULL);
    cpl_ensure(inum < imlist->ni, CPL_ERROR_ACCESS_OUT_OF_RANGE, NULL);

    return imlist->images[inum] ;
}

/*----------------------------------------------------------------------------*/
/**
  @brief    Get an image from a list of images
  @param    imlist    the image list
  @param    inum    the image id (from 0 to number of images-1)
  @return   A pointer to the image or NULL in error case.
  @see cpl_imagelist_get
 */
/*----------------------------------------------------------------------------*/
const cpl_image * cpl_imagelist_get_const(const cpl_imagelist * imlist,
                                    int                   inum)
{
    return cpl_imagelist_get((cpl_imagelist *)imlist, inum);
}

/*----------------------------------------------------------------------------*/
/**
  @brief    Insert an image into an imagelist
  @param    imlist   The imagelist
  @param    im       The image to insert
  @param    pos      The list position (from 0 to number of images)
  @return   CPL_ERROR_NONE or the relevant #_cpl_error_code_ on error

  The image is put at the position pos in the image list. The image that is
  already there is deallocated.

  It is allowed to specify the position equal to the number of images in the
  list. This will increment the size of the imagelist.

  No action occurs if an image is inserted more than once into the same
  position. It is not allowed to insert the same image into two different
  positions in a list.

  It is not allowed to insert images of different size into a list.

  You should not deallocate the image after it has been inserted into an
  imagelist, since the cpl_imagelist_delete() will deallocate its images.

  Possible #_cpl_error_code_ set in this function:
  - CPL_ERROR_NULL_INPUT if (one of) the input pointer(s) is NULL
  - CPL_ERROR_ILLEGAL_INPUT if pos is negative
  - CPL_ERROR_TYPE_MISMATCH if im and imlist are of different types
  - CPL_ERROR_INCOMPATIBLE_INPUT if im and imlist have different sizes
  - CPL_ERROR_ACCESS_OUT_OF_RANGE if pos is bigger than the number of
    images in imlist
 */
/*----------------------------------------------------------------------------*/
cpl_error_code cpl_imagelist_set(
        cpl_imagelist * imlist,
        cpl_image     * im,
        int             pos)
{
    int                 i;

    cpl_ensure_code(imlist,            CPL_ERROR_NULL_INPUT);
    cpl_ensure_code(im,                CPL_ERROR_NULL_INPUT);
    cpl_ensure_code(pos >= 0,          CPL_ERROR_ILLEGAL_INPUT);
    cpl_ensure_code(pos <= imlist->ni, CPL_ERROR_ACCESS_OUT_OF_RANGE);

    /* Do nothing if the image is already there */
    if (pos < imlist->ni && im == imlist->images[pos]) return CPL_ERROR_NONE;

    /* Do not allow the image to be inserted into two different positions */
    for (i = 0; i < imlist->ni; i++)
        cpl_ensure_code(im != imlist->images[i], CPL_ERROR_ILLEGAL_INPUT);

    if (pos > 0 || imlist->ni > 0) {
        /* Require images to have the same size and type */
        cpl_ensure_code(cpl_image_get_size_x(im) ==
                        cpl_image_get_size_x(imlist->images[0]),
                        CPL_ERROR_INCOMPATIBLE_INPUT);
        cpl_ensure_code(cpl_image_get_size_y(im) ==
                        cpl_image_get_size_y(imlist->images[0]),
                        CPL_ERROR_INCOMPATIBLE_INPUT);
        cpl_ensure_code(cpl_image_get_type(im) ==
                        cpl_image_get_type(imlist->images[0]),
                        CPL_ERROR_TYPE_MISMATCH);
    }

    if (pos == imlist->ni) {
        imlist->ni++;
        imlist->images = cpl_realloc(imlist->images,
                                 imlist->ni * sizeof(cpl_image*));
    } else {
        /* Delete any existing image at that position */
        cpl_image_delete(imlist->images[pos]);
    }

    imlist->images[pos] = im ;

    return CPL_ERROR_NONE ;
}

/*----------------------------------------------------------------------------*/
/**
  @brief    Remove an image from an imagelist
  @param    self   The imagelist
  @param    pos      The list position (from 0 to number of images-1)
  @return   The pointer to the removed image or NULL in error case

  The specified image is not deallocated, it is simply removed from the
  list. The pointer to the image is returned to let the user decide to
  deallocate it or not.
  Eventually, the image will have to be deallocated with cpl_image_delete().

  Possible #_cpl_error_code_ set in this function:
  - CPL_ERROR_NULL_INPUT if (one of) the input pointer(s) is NULL
  - CPL_ERROR_ILLEGAL_INPUT if pos is negative
  - CPL_ERROR_ACCESS_OUT_OF_RANGE if pos is bigger than the number of
    images in self
 */
/*----------------------------------------------------------------------------*/
cpl_image * cpl_imagelist_unset(cpl_imagelist * self,
                                int             pos)
{
    cpl_image * out;
    int         i;

    cpl_ensure(self, CPL_ERROR_NULL_INPUT, NULL) ;
    cpl_ensure(pos >= 0, CPL_ERROR_ILLEGAL_INPUT, NULL) ;
    cpl_ensure(pos < self->ni, CPL_ERROR_ACCESS_OUT_OF_RANGE, NULL) ;

    /* Get pointer to image to be removed */
    out = self->images[pos];

    /* Move the following images one position towards zero */
    for (i=pos + 1; i < self->ni; i++) {
        self->images[i-1] = self->images[i];
    }

    /* Decrement of the size */
    self->ni--;

    return out;
}

/*----------------------------------------------------------------------------*/
/**
  @brief    Free memory associated to a cpl_imagelist object.
  @param    imlist    the image list

  @return   Nothing

   The function deallocates the memory used by the imagelist @em imlist.
   If @em imlist is @c NULL, nothing is done, and no error is set.

 */
/*----------------------------------------------------------------------------*/
void cpl_imagelist_delete(cpl_imagelist * imlist)
{
    int            i ;

    if (imlist == NULL) return ;
    for (i=0 ; i<imlist->ni ; i++) cpl_image_delete(imlist->images[i]) ;
    if (imlist->images) cpl_free(imlist->images);
    cpl_free(imlist) ;
    return ;
}

/*----------------------------------------------------------------------------*/
/**
  @brief    Copy an image list
  @param    imlist    Source image list.
  @return   1 newly allocated image list, or NULL on error.

  Copy an image list into a new image list object.
  The returned image list must be deallocated using cpl_imagelist_delete().

  Possible #_cpl_error_code_ set in this function:
  - CPL_ERROR_NULL_INPUT if (one of) the input pointer(s) is NULL
 */
/*----------------------------------------------------------------------------*/
cpl_imagelist * cpl_imagelist_duplicate(const cpl_imagelist * imlist)
{
    cpl_imagelist *   out ;
    int               i;

    cpl_ensure(imlist != NULL, CPL_ERROR_NULL_INPUT, NULL) ;

    /* Create the new imagelist */
    out = cpl_imagelist_new() ;

    /* Duplicate the images */
    for (i=0 ; i<imlist->ni ; i++) {
        cpl_imagelist_set(out, cpl_image_duplicate(imlist->images[i]), i) ;
    }

    return out ;
}

/*----------------------------------------------------------------------------*/
/**
  @brief    Reject one or more images in a list according to an array of flags.
  @param    imlist  Non-empty imagelist to examine for image rejection.
  @param    valid   Vector of flags (>=-0.5: valid, <-0.5: invalid)
  @return   CPL_ERROR_NONE or the relevant #_cpl_error_code_ on error

  This function takes an imagelist and a vector of flags. The imagelist and
  vector must have equal lengths.

  Images flagged as invalid are removed from the list.

  The removal of image(s) will reduce the length of the list accordingly.

  Possible #_cpl_error_code_ set in this function:
  - CPL_ERROR_NULL_INPUT if (one of) the input pointer(s) is NULL
  - CPL_ERROR_INCOMPATIBLE_INPUT if the vector size and the image list
    size are different
 */
/*----------------------------------------------------------------------------*/
cpl_error_code cpl_imagelist_erase(
        cpl_imagelist    * imlist,
        const cpl_vector * valid)
{
    int          nkeep = 0;
    int          i;

    /* Check entries */
    cpl_ensure_code(imlist, CPL_ERROR_NULL_INPUT);
    cpl_ensure_code(valid,  CPL_ERROR_NULL_INPUT);
    cpl_ensure_code(cpl_vector_get_size(valid) == imlist->ni,
                    CPL_ERROR_INCOMPATIBLE_INPUT);

    for (i=0 ; i < imlist->ni ; i++) {
        if (cpl_vector_get(valid, i) >= -0.5) {
            /* image is to be kept, place it in the 1st free position */
            imlist->images[nkeep] = imlist->images[i];
            nkeep++;
        } else {
            /* image is to be erased, delete it */
            cpl_image_delete(imlist->images[i]);
        }
    }

    /* Update the size of the altered list */
    imlist->ni = nkeep;

    return CPL_ERROR_NONE ;
}

/*----------------------------------------------------------------------------*/
/**
  @brief    Save an imagelist to disk in FITS format.
  @param    imlist      Imagelist to save
  @param    filename    Name of the FITS file to write
  @param    bpp         Bits per pixels
  @param    pl          Property list for the output header or NULL
  @param    mode        The desired output options (combined with bitwise or)
  @return   the #_cpl_error_code_ or CPL_ERROR_NONE
  @see cpl_image_save()

  This function saves an image list to a FITS file, using cfitsio. If a
  property list is provided, it is written to the named file before the
  pixels are written.

  The requested pixel depth (bpp) follows the FITS convention. Possible values
  are CPL_BPP_8_UNSIGNED (8), CPL_BPP_16_SIGNED (16), CPL_BPP_16_UNSIGNED (-16),
  CPL_BPP_32_SIGNED (32), CPL_BPP_IEEE_FLOAT (-32), CPL_BPP_IEEE_DOUBLE (-64)

  In the case of CPL_BPP_16_UNSIGNED, the BITPIX stays 16, the data are
  stored as signed (32768 is subtracted to the images) and BZERO is set to
  32768 as specified by the FITS standard.

  Supported image lists types are CPL_TYPE_DOUBLE, CPL_TYPE_FLOAT, CPL_TYPE_INT.

  Supported output modes are CPL_IO_CREATE (create a new file) and
  CPL_IO_EXTEND  (append to an existing file)

  If you are in append mode, make sure that the file has writing
  permissions. You may have problems if you create a file in your
  application and append something to it with the umask set to 222. In
  this case, the file created by your application would not be writable,
  and the append would fail.

  Possible #_cpl_error_code_ set in this function:
  - CPL_ERROR_NULL_INPUT if (one of) the input pointer(s) is NULL
  - CPL_ERROR_ILLEGAL_INPUT if the bpp value or the mode is not supported
  - CPL_ERROR_FILE_IO if the file cannot be written
  - CPL_ERROR_INVALID_TYPE if the passed image list type is not supported
 */
/*----------------------------------------------------------------------------*/
cpl_error_code cpl_imagelist_save(
        const cpl_imagelist     *   imlist,
        const char              *   filename,
        cpl_type_bpp                bpp,
        const cpl_propertylist  *   pl,
        unsigned                    mode)
{
    int                 nx, ny ;
    char            *   sval ;
    int                 fio_status=0 ;
    fitsfile        *   fptr ;
    long                naxes[3] ;
    long                fpixel[3] ;
    double              dval ;
    int                 i ;
    const char       * badkeys = mode & CPL_IO_EXTEND ?
	CPL_FITS_BADKEYS_EXT  "|" CPL_FITS_COMPRKEYS :
	CPL_FITS_BADKEYS_PRIM "|" CPL_FITS_COMPRKEYS ;
    int cfitsiotype ;
    const cpl_image       *   first = cpl_imagelist_get_const(imlist, 0);
    cpl_type type = cpl_image_get_type(first);

    /* Test entries */
    cpl_ensure_code(filename, CPL_ERROR_NULL_INPUT);
    cpl_ensure_code(imlist, CPL_ERROR_NULL_INPUT);
    cpl_ensure_code(bpp==CPL_BPP_8_UNSIGNED  || bpp==CPL_BPP_16_SIGNED ||
                    bpp==CPL_BPP_16_UNSIGNED || bpp==CPL_BPP_32_SIGNED ||
                    bpp==CPL_BPP_IEEE_FLOAT  || bpp==CPL_BPP_IEEE_DOUBLE,
                    CPL_ERROR_ILLEGAL_INPUT);
    cpl_ensure_code(mode <= CPL_IO_MAX, CPL_ERROR_ILLEGAL_INPUT) ;

    cpl_ensure_code((mode & CPL_IO_CREATE) ^ (mode & CPL_IO_EXTEND),
		    CPL_ERROR_ILLEGAL_INPUT);

    cpl_ensure_code(cpl_imagelist_is_uniform(imlist)==0,
                    CPL_ERROR_ILLEGAL_INPUT);
    nx = cpl_image_get_size_x(cpl_imagelist_get_const(imlist, 0));
    ny = cpl_image_get_size_y(cpl_imagelist_get_const(imlist, 0));

    cpl_ensure_code(type == CPL_TYPE_DOUBLE ||
		    type == CPL_TYPE_FLOAT ||
		    type == CPL_TYPE_INT, CPL_ERROR_INVALID_TYPE);

    if (mode & CPL_IO_EXTEND) {
	/* Open the file */
	fits_open_diskfile(&fptr, filename, READWRITE, &fio_status);
	if (fio_status != 0) {
	    return cpl_error_set_message_macro
		(cpl_func, CPL_ERROR_FILE_IO, __FILE__, __LINE__,
		 "fits_open_diskfile('%s') returned %d: %s", filename,
		 fio_status, cpl_tools_get_cfitsio_msg(fio_status));
	}
    } else {
	/* Create the file */
	sval = cpl_sprintf("!%s", filename) ;
	fits_create_file(&fptr, sval, &fio_status);
	cpl_free(sval) ;
	if (fio_status != 0) {
	    fits_close_file(fptr, &fio_status) ;
	    cpl_ensure_code(0, CPL_ERROR_FILE_IO) ;
	}
    }

    /* Create the imagelist in the primary unit */
    naxes[0] = nx ;
    naxes[1] = ny ;
    naxes[2] = (long)(imlist->ni) ;
    fits_create_img(fptr, bpp, 3, naxes, &fio_status);

    /* Add BSCALE DATE */
    dval = 1.0 ;
    /* FIXME: Cast string literal to avoid compiler warnings
       - and pray that CFITSIO does not try to modify them :-(((( */
    fits_write_key(fptr, TDOUBLE, (char*)"BSCALE", &dval,
		   (char*)"Scale factor", &fio_status);

    if (mode & CPL_IO_CREATE)
	fits_write_date(fptr, &fio_status);

    /* Check  */
    if (fio_status != 0) {
        fits_close_file(fptr, &fio_status) ;
        cpl_ensure_code(0, CPL_ERROR_FILE_IO) ;
    }

    /* Add the property list */
    if (cpl_fits_add_properties(fptr, pl, badkeys)!=CPL_ERROR_NONE) {
        fits_close_file(fptr, &fio_status) ;
        cpl_ensure_code(0, CPL_ERROR_ILLEGAL_INPUT) ;
    }

    cfitsiotype = type == CPL_TYPE_DOUBLE ? TDOUBLE
	: (type == CPL_TYPE_FLOAT ? TFLOAT : TINT);

    /* Loop on the images */
    fpixel[0] = fpixel[1] = 1 ;
    for (i=0 ; i<imlist->ni ; i++) {
	const cpl_image * image = cpl_imagelist_get_const(imlist, i);
	const void      * data  = cpl_image_get_data_const(image);

        fpixel[2] = i+1 ;

	/* Write the pixels */
	fits_write_pix(fptr, cfitsiotype, fpixel, nx*ny,
		       (void*)data, &fio_status);
    }

    /* Check  */
    if (fio_status != 0) {
        fits_close_file(fptr, &fio_status) ;
        return cpl_error_set_message_macro
            (cpl_func, CPL_ERROR_FILE_IO,
             __FILE__, __LINE__, "CFITSIO error (status=%d, filename='%s', "
             "bpp=%d, mode=%d): %s", fio_status, filename, bpp, mode,
             cpl_tools_get_cfitsio_msg(fio_status));
    }

    /* Close and write on disk */
    fits_close_file(fptr, &fio_status) ;

    return CPL_ERROR_NONE ;
}

/*----------------------------------------------------------------------------*/
/**
  @brief    Determine if an imagelist contains images of equal size and type
  @param    imlist    The imagelist to check
  @return   Zero if uniform, positive if non-uniform and negative on error.

  The function returns 1 if the list is empty.

  Possible #_cpl_error_code_ set in this function:
  - CPL_ERROR_NULL_INPUT if (one of) the input pointer(s) is NULL
 */
/*----------------------------------------------------------------------------*/
int cpl_imagelist_is_uniform(const cpl_imagelist * imlist)
{
    cpl_type     type;
    int          nx, ny;
    int          i;

    cpl_ensure(imlist != NULL, CPL_ERROR_NULL_INPUT, -1);
    if (imlist->ni == 0) return 1;

    /* Check the images */
    nx   = cpl_image_get_size_x(imlist->images[0]);
    ny   = cpl_image_get_size_y(imlist->images[0]);
    type = cpl_image_get_type  (imlist->images[0]);

    for (i=1 ; i < imlist->ni ; i++) {
        if (cpl_image_get_size_x(imlist->images[i]) != nx)   return i+1;
        if (cpl_image_get_size_y(imlist->images[i]) != ny)   return i+1;
        if (cpl_image_get_type  (imlist->images[i]) != type) return i+1;
    }
    return 0 ;
}

/*----------------------------------------------------------------------------*/
/**
  @brief    Dump structural information of images in an imagelist
  @param    self    Imagelist to dump
  @param    stream  Output stream, accepts @c stdout or @c stderr
  @return   CPL_ERROR_NONE or the relevant #_cpl_error_code_ on error

  Possible #_cpl_error_code_ set in this function:
  - CPL_ERROR_NULL_INPUT if an input pointer is NULL
  - CPL_ERROR_FILE_IO if a write operation fails
 */
/*----------------------------------------------------------------------------*/
cpl_error_code cpl_imagelist_dump_structure(const cpl_imagelist * self,
					    FILE * stream)
{
    const char * msg    = "Imagelist with %d image(s)\n";
    const int    msgmin = strlen(msg) - 5;

    int i;

    cpl_ensure_code(self   != NULL, CPL_ERROR_NULL_INPUT);
    cpl_ensure_code(stream != NULL, CPL_ERROR_NULL_INPUT);

    cpl_ensure_code( fprintf(stream,  msg, self->ni) >= msgmin,
		     CPL_ERROR_FILE_IO );

    for (i = 0; i < self -> ni; i++) {
	const cpl_image * image   = cpl_imagelist_get_const(self, i);
	const char      * imsg    = "Image nb %d of %d in imagelist\n";
	const int         imsgmin = strlen(imsg) - 5;

	cpl_ensure_code( fprintf(stream,  imsg, i, self->ni) >= imsgmin,
			 CPL_ERROR_FILE_IO );

	cpl_ensure_code( cpl_image_dump_structure(image, stream),
			 cpl_error_get_code() );
    }

    return CPL_ERROR_NONE;
}

/*----------------------------------------------------------------------------*/
/**
  @brief    Dump pixel values of images in a CPL imagelist
  @param    self    Imagelist to dump
  @param    llx     Specifies the window position
  @param    lly     Specifies the window position
  @param    urx     Specifies the window position
  @param    ury     Specifies the window position
  @param    stream  Output stream, accepts @c stdout or @c stderr
  @return   CPL_ERROR_NONE or the relevant #_cpl_error_code_ on error

  Possible #_cpl_error_code_ set in this function:
  - CPL_ERROR_NULL_INPUT if an input pointer is NULL
  - CPL_ERROR_FILE_IO if a write operation fails
  - CPL_ERROR_ACCESS_OUT_OF_RANGE if the defined window is not in the image
  - CPL_ERROR_ILLEGAL_INPUT if the window definition is wrong (e.g llx > urx)
 */
/*----------------------------------------------------------------------------*/
cpl_error_code cpl_imagelist_dump_window(const cpl_imagelist * self,
					 int llx, int lly, int urx, int ury,
					 FILE * stream)
{
    int i;

    cpl_ensure_code(self   != NULL, CPL_ERROR_NULL_INPUT);
    cpl_ensure_code(stream != NULL, CPL_ERROR_NULL_INPUT);

    for (i = 0; i < self -> ni; i++) {
	const cpl_image * image   = cpl_imagelist_get_const(self, i);
	const char      * imsg    = "Image nb %d of %d in imagelist\n";
	const int         imsgmin = strlen(imsg) - 5;

	cpl_ensure_code( fprintf(stream,  imsg, i, self->ni) >= imsgmin,
			 CPL_ERROR_FILE_IO );

	cpl_ensure_code( cpl_image_dump_window(image, llx, lly, urx, ury,
					       stream),
			 cpl_error_get_code() );
    }

    return CPL_ERROR_NONE;
}

/**@}*/

#define CPL_OPERATION CPL_IMLIST_IO_LOAD_WIN
/*----------------------------------------------------------------------------*/
/*
  @brief    Internal loading function, wrapped by cpl_imagelist_load{, _window}().
 */
/*----------------------------------------------------------------------------*/
static cpl_imagelist * cpl_imagelist_load_internal(const char  *   filename,
						   cpl_type        type,
						   int             extnum,
						   int             llx,
						   int             lly,
						   int             urx,
						   int             ury)
{
    fitsfile        *   fptr ;
    int                 fio_status=0 ;
    int                 naxis ;
    long            *   naxes ;
    int                 nplanes ;
    cpl_imagelist   *   imlist;
    long            *   fpixel ;
    long            *   inc ;
    long            *   lpixel ;
    void            *   pimage ;
    cpl_image       *   image = NULL;
    int                 i ;

    /* Test entries */
    cpl_ensure(filename != NULL,   CPL_ERROR_NULL_INPUT, NULL);
    cpl_ensure(extnum   >= 0,      CPL_ERROR_ILLEGAL_INPUT, NULL);

    /* Open the file and move to extension */
    fits_open_diskfile(&fptr, filename, READONLY, &fio_status) ;
    if (fio_status != 0) {
        (void)cpl_error_set_message_macro(cpl_func, CPL_ERROR_FILE_IO,
                                          __FILE__, __LINE__, "filename='%s', "
                                          "xtnum=%d, fits_open_diskfile() "
                                          "returned %d: %s", filename,
                                          extnum, fio_status,
                                          cpl_tools_get_cfitsio_msg(fio_status));
        return NULL;
    }

    if (extnum > 0) fits_movabs_hdu(fptr, extnum + 1, NULL, &fio_status) ;
    if (fio_status != 0) {
        int fio_status2 = 0;
        fits_close_file(fptr, &fio_status2);
        (void)cpl_error_set_message_macro(cpl_func, CPL_ERROR_FILE_IO,
                                          __FILE__, __LINE__, "filename='%s', "
                                          "xtnum=%d, fits_movabs_hdu() "
                                          "returned %d: %s", filename,
                                          extnum, fio_status,
                                          cpl_tools_get_cfitsio_msg(fio_status));
        return NULL;
    }

    /* Get the file dimension */
    fits_get_img_dim(fptr, &naxis, &fio_status) ;
    if (naxis != 2 && naxis != 3) {
        fits_close_file(fptr, &fio_status) ;
        cpl_ensure(0, CPL_ERROR_ILLEGAL_INPUT, NULL) ;
    }

    /* Get the file size */
    naxes = cpl_malloc(naxis * sizeof(long)) ;
    fits_get_img_size(fptr, naxis, naxes, &fio_status) ;
    if (naxis == 2)     nplanes = 1 ;
    else                nplanes = naxes[2] ;

    /* Check the zones */
    if (llx == -1 && lly == -1 && urx == -1 && ury == -1) {
	llx = lly = 1;
	urx = naxes[0];
	ury = naxes[1];
    } else if (llx < 1 || lly < 1 || urx > naxes[0] || ury > naxes[1] ||
            urx<llx || ury < lly) {
        fits_close_file(fptr, &fio_status) ;
        cpl_free(naxes) ;
        cpl_ensure(0, CPL_ERROR_ILLEGAL_INPUT, NULL) ;
    }
    cpl_free(naxes) ;

    /* Create the image list */
    imlist = cpl_imagelist_new() ;

    /* Create fpixel */
    fpixel = cpl_malloc(naxis * sizeof(long)) ;
    lpixel = cpl_malloc(naxis * sizeof(long)) ;
    inc = cpl_malloc(naxis * sizeof(long)) ;
    for (i=0 ; i<naxis ; i++) inc[i] = 1 ;
    fpixel[0] = llx ;
    fpixel[1] = lly ;
    lpixel[0] = urx ;
    lpixel[1] = ury ;

    /* Switch on the requested image type */
    switch (type) {
    case CPL_TYPE_DOUBLE: {
#define CPL_CLASS CPL_CLASS_DOUBLE
#include "cpl_imagelist_io_body.h"
#undef CPL_CLASS
    }
    break ;

    case CPL_TYPE_FLOAT: {
#define CPL_CLASS CPL_CLASS_FLOAT
#include "cpl_imagelist_io_body.h"
#undef CPL_CLASS
    }
    break ;
    case CPL_TYPE_INT: {
#define CPL_CLASS CPL_CLASS_INT
#include "cpl_imagelist_io_body.h"
#undef CPL_CLASS
    }
    break ;

    case CPL_TYPE_UNSPECIFIED: {
	cpl_type outtype;

	/*
	 * 0 is an invalid FITS type value. In this way, we make sure to
	 * fall into the 'default' branch' of the switch
	 */
	int cfitsiotype, filetype = 0;

	fits_get_img_type(fptr, &filetype, &fio_status);

	switch (filetype) {
	case BYTE_IMG    :
	case SHORT_IMG   :
	case LONG_IMG    :
	case LONGLONG_IMG:
	    outtype = CPL_TYPE_INT;
	    cfitsiotype = TINT;
	    break;
	case FLOAT_IMG   :
	    outtype = CPL_TYPE_FLOAT;
	    cfitsiotype = TFLOAT;
	    break;
	case DOUBLE_IMG  :
	    outtype = CPL_TYPE_DOUBLE;
	    cfitsiotype = TDOUBLE;
	    break;
	default:
	    outtype = CPL_TYPE_INVALID;
	    break;

	}

	if (outtype != CPL_TYPE_INVALID) {
	    for (i=0 ; i < nplanes ; i++) {
		image = cpl_image_new(urx - llx + 1, ury - lly + 1, outtype);
		pimage = cpl_image_get_data(image) ;
		if (naxis == 3) fpixel[2] = lpixel[2] = i+1 ;
		fits_read_subset(fptr, cfitsiotype, fpixel, lpixel, inc,
				 NULL, pimage, NULL, &fio_status) ;
		if (cpl_imagelist_set(imlist, image, i)) {
		    cpl_image_delete(image);
		    break;
		}
		image = NULL;
	    }
	}

	break;
    }

    default:
        fits_close_file(fptr, &fio_status) ;
        cpl_free(fpixel) ;
        cpl_free(lpixel) ;
        cpl_free(inc) ;
        cpl_imagelist_delete(imlist);
        cpl_ensure(0, CPL_ERROR_INVALID_TYPE, NULL);
        break ;
    }
    cpl_free(fpixel) ;
    cpl_free(lpixel) ;
    cpl_free(inc) ;

    if (imlist->ni < nplanes) {
        cpl_imagelist_delete(imlist);
        fits_close_file(fptr, &fio_status) ;
        cpl_ensure(0, CPL_ERROR_ILLEGAL_INPUT, NULL) ;
    }

    /* Return */
    fits_close_file(fptr, &fio_status) ;
    return imlist ;
}
#undef CPL_OPERATION
